#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       drawingarea.py
#       
#       Copyright 2010-2011 Sandro Mani <manisandro@gmail.com>
#       
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import cairo
import poppler
import Image
import ImageFilter
import ImageEnhance
import array
import gtk
import gobject
import os
import re
import urllib
import gettext
import threadtools
_ = gettext.gettext

class dotdict(dict):
	""" A dict class which allows member access by dot syntax. """
	def __getattr__(self, attr):
		return self.get(attr, None)
	__setattr__= dict.__setitem__
	__delattr__= dict.__delitem__

class Drawingarea:
	""" This class is responsible for the drawingarea section of the program. """
	def __init__(self, builder, main):
		""" Constructor. """
		self.main=main
		# Widgets
		self.tb_bestzoom=builder.get_object("tb_bestzoom")
		self.tb_resetzoom=builder.get_object("tb_resetzoom")
		self.tb_zoomin=builder.get_object("tb_zoomin")
		self.tb_zoomout=builder.get_object("tb_zoomout")
		self.tb_toggle_controls=builder.get_object("tb_toggle_controls")
		self.tb_doocr=builder.get_object("tb_doocr")
		
		self.menu_doocr=gtk.Menu()
		self.menu_doocr_current=gtk.MenuItem(_("Current Page"))
		self.menu_doocr.append(self.menu_doocr_current)
		self.menu_doocr_all=gtk.MenuItem(_("All Pages"))
		self.menu_doocr.append(self.menu_doocr_all)
		self.menu_doocr_pages=gtk.MenuItem(_("Pages..."))
		self.menu_doocr.append(self.menu_doocr_pages)
		self.menu_doocr.show_all()
		self.dialog_pages=builder.get_object("dialog_pages")
		self.entry_pages=builder.get_object("entry_pages")
		
		self.menu_bestzoom=builder.get_object("menu_bestzoom")
		self.menu_resetzoom=builder.get_object("menu_resetzoom")
		self.menu_zoomin=builder.get_object("menu_zoomin")
		self.menu_zoomout=builder.get_object("menu_zoomout")
		self.menu_toggle_controls=builder.get_object("menu_toggle_controls")
		
		self.layout=builder.get_object("layout")
		self.scrolledwindow=builder.get_object("scrolledwindow")
		self.hscrollbar=self.scrolledwindow.get_hscrollbar()
		self.vscrollbar=self.scrolledwindow.get_vscrollbar()
		self.hadjustment=self.scrolledwindow.get_hadjustment()
		self.vadjustment=self.scrolledwindow.get_vadjustment()
		
		self.toolbar_navigation=builder.get_object("toolbar_navigation")
		self.spin_navigation_page=builder.get_object("spinbutton_navigation_page")
		self.label_navigation_totpages=builder.get_object("label_navigation_totpages")
		
		self.toolbar_controls=builder.get_object("toolbar_controls")
		self.label_controls_dpi=builder.get_object("label_controls_dpi")
		self.spin_controls_resolution=builder.get_object("spinbutton_controls_resolution")
		self.spin_controls_resolution.set_value(200)
		self.spin_controls_brightness=builder.get_object("spinbutton_controls_brightness")
		self.spin_controls_contrast=builder.get_object("spinbutton_controls_contrast")
		
		# Canvas    (x,y): top left corner, (w,h): dimensions at 1x zoom, (ow,oh): dimensions without digital rescaling, (cw, ch): current dimensions, s: scaling factor. a: angle
		self.canvas_geo=dotdict({'x':0,'y':0,'w':0,'h':0,'ow':0,'oh':0,'cw':0,'ch':0,'s':1.0,'a':0})
		self.canvas=gtk.DrawingArea()
		self.canvas.modify_bg(gtk.STATE_NORMAL,  gtk.gdk.Color(0.5, 0.5, 0.5))
		self.canvas.set_size_request(self.canvas_geo.w, self.canvas_geo.h)
		self.canvas.set_redraw_on_allocate(False)
		self.layout.put(self.canvas, self.canvas_geo.x, self.canvas_geo.y)
		self.zoom_fit=False
		self.blur_timer=-1
		self.controls_timer=-1
		self.image=None
		self.document=None
		self.PILimage=None
		
		# Selection
		self.selection=None
		self.wsel=None		#Working selection (only used during dragging)
		self.ssel=None		#Scaled selection, i.e. working selection but has w>0 and h>0
		self.nsel=None		#Normal selection, i.e. selection at 100% zoom, with w>0 and h>0
		self.selcolor=[0.0, 0.0, 1.0]
		self.__selection_set_color()
		
		# Layout
		self.scroll_animation=False
		self.layout_scrollx=0
		self.layout_scrolly=0
		self.layout_width=0
		self.layout_height=0
		self.layout.add_events(gtk.gdk.POINTER_MOTION_MASK | gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK)
		self.layout.modify_bg(gtk.STATE_NORMAL, gtk.gdk.Color(0.5, 0.5, 0.5))
		
		# Signals
		self.scrolledwindow.connect("size_allocate", self.__scrolledwindow_resized)
		self.layout.connect("scroll-event", self.__layout_scrolled)
		self.canvas.connect("expose-event", self.__canvas_exposed)
		self.canvas.connect("style_set",self.__selection_set_color)
		self.layout.connect("button_press_event", self.__selection_start)
		builder.get_object("tb_zoomin").connect("clicked", self.set_zoom,"+")
		builder.get_object("tb_zoomout").connect("clicked", self.set_zoom,"-")
		builder.get_object("tb_bestzoom").connect("clicked", self.set_zoom,"f")
		builder.get_object("tb_resetzoom").connect("clicked", self.set_zoom,1.0)
		builder.get_object("tb_lrotate").connect("clicked", self.rotate,+90)
		builder.get_object("tb_rrotate").connect("clicked", self.rotate,-90)
		self.tb_doocr.connect("clicked", self.__recognize)
		self.menu_doocr_current.connect("activate", lambda w: self.main.recognize())
		self.menu_doocr_all.connect("activate", self.__recognize_all)
		self.menu_doocr_pages.connect("activate", self.__recognize_pages)
		builder.get_object("menu_zoomin").connect("activate", self.set_zoom,"+")
		builder.get_object("menu_zoomout").connect("activate", self.set_zoom,"-")
		builder.get_object("menu_bestzoom").connect("activate", self.set_zoom,"f")
		builder.get_object("menu_resetzoom").connect("activate", self.set_zoom,1.0)
		builder.get_object("menu_lrotate").connect("activate", self.rotate,+90)
		builder.get_object("menu_rrotate").connect("activate", self.rotate,-90)
		self.menu_toggle_controls.connect("toggled", self.__toggle_controls)
		self.tb_toggle_controls.connect("toggled", self.__toggle_controls)
		self.spin_navigation_page.connect("value-changed", lambda w: self.__render_page(w.get_value_as_int()))
		self.spin_controls_resolution.connect("value-changed", self.__resolution_changed)
		self.spin_controls_brightness.connect("value-changed", self.__brightness_contrast_changed)
		self.spin_controls_contrast.connect("value-changed", self.__brightness_contrast_changed)
		self.entry_pages.connect("activate", lambda w: self.dialog_pages.response(gtk.RESPONSE_OK))

	### CANVAS ###	
	def __scrolledwindow_resized(self, widget, event):
		""" Called when the scrolled window containing the canvas is resized. """
		if self.layout.allocation.width!=self.layout_width or self.layout.allocation.height!=self.layout_height:
			self.layout_width=self.layout.allocation.width
			self.layout_height=self.layout.allocation.height
			if self.image==None:
				return
			elif self.zoom_fit:
				self.set_zoom(None,"f")
			else:
				self.__position_canvas()
	
	def __layout_scrolled(self, widget, event):
		""" Called when scroll events are captured on the layout widget. """
		if event.state&gtk.gdk.CONTROL_MASK==gtk.gdk.CONTROL_MASK:
			if event.direction==gtk.gdk.SCROLL_UP and self.canvas_geo.s*1.25<=10:
				self.set_zoom(None,"+",[float(event.x-self.canvas_geo.x)/self.canvas_geo.cw, float(event.y-self.canvas_geo.y)/self.canvas_geo.ch])
			elif event.direction==gtk.gdk.SCROLL_DOWN and self.canvas_geo.s*0.8>=0.1:
				self.set_zoom(None,"-",[float(event.x-self.canvas_geo.x)/self.canvas_geo.cw, float(event.y-self.canvas_geo.y)/self.canvas_geo.ch])
			return True
		else:
			return False
	
	def __canvas_exposed(self, widget, event):
		""" Called when the canvas is to be repainted. """
		gobject.source_remove(self.blur_timer)
		self.layout_scrollx=self.hadjustment.get_value()
		self.layout_scrolly=self.vadjustment.get_value()
		if self.image!=None:
			context=self.canvas.window.cairo_create()
			context.save()
			# Area (not affected by transformation, i.e. view coordinates)
			context.rectangle(event.area)
			# Set up transformations
			context.translate(self.canvas_geo.cw/2., self.canvas_geo.ch/2.)
			context.scale(self.canvas_geo.s, self.canvas_geo.s)
			context.rotate(self.canvas_geo.a/180.0*3.141592654)
			context.translate(-self.canvas_geo.w/2., -self.canvas_geo.h/2.)
			# Set source and apply all transformations to it (which are specified in reverse order)
			context.set_source_surface(self.image)
			context.fill()
			context.restore()
			self.__selection_draw(context, event.area)
			if self.canvas_geo.s<1 and self.main.get_config().get_antialiasing():
				self.blur_timer=gobject.timeout_add(100, threadtools.SubprocessThread(self.__calc_downsampled, self.__draw_downsampled, event.area).start)
	
	def __calc_downsampled(self, area):
		if self.image==None: return [False]
		w, h=self.image.get_width(), self.image.get_height()
		image=Image.frombuffer("RGBA", (w, h), self.image.get_data(),"raw","RGBA", 0, 1)
		s=self.canvas_geo.s
		w=int(round(w*s))
		h=int(round(h*s))
		image=image.resize((w, h), Image.ANTIALIAS)
		surface=cairo.ImageSurface.create_for_data(array.array('B', image.tostring()), cairo.FORMAT_ARGB32, w, h, w*4)
		return [True, area, surface, w, h, s]
	
	def __draw_downsampled(self, res):
		if res[0]==True and res[5]==self.canvas_geo.s:
			context=self.canvas.window.cairo_create()
			context.save()
			# Area (not affected by transformation, i.e. view coordinates)
			context.rectangle(res[1])
			# Set up transformations
			context.translate(self.canvas_geo.cw/2., self.canvas_geo.ch/2.)
			context.rotate(self.canvas_geo.a/180.0*3.141592654)
			context.translate(-res[3]/2., -res[4]/2.)
			# Set source and apply all transformations to it (which are specified in reverse order)
			context.set_source_surface(res[2])
			context.fill()
			context.restore()
			self.__selection_draw(context, res[1])
	
	def __position_canvas(self, adjustments=None):
		""" Positions the canvas in order to center the image in the available area. """
		if not self.canvas.window: return # canvas is not visible
		if self.canvas_geo.cw<=self.layout_width:
			self.canvas_geo.x=int(round((self.layout_width-self.canvas_geo.cw)/2))
			lw=self.layout_width
		else:
			self.canvas_geo.x=0
			lw=self.canvas_geo.cw
		if self.canvas_geo.ch<=self.layout_height:
			self.canvas_geo.y=int(round((self.layout_height-self.canvas_geo.ch)/2))
			lh=self.layout_height
		else:
			self.canvas_geo.y=0
			lh=self.canvas_geo.ch
		self.canvas.window.freeze_updates()
		self.layout.set_size(lw, lh)
		self.layout.move(self.canvas, self.canvas_geo.x, self.canvas_geo.y)
		if adjustments!=None:
			self.hscrollbar.set_value(adjustments[0]*(self.hadjustment.get_upper()-self.hadjustment.get_page_size()))
			self.vscrollbar.set_value(adjustments[1]*(self.vadjustment.get_upper()-self.vadjustment.get_page_size()))
		self.canvas.window.thaw_updates()
	
	def set_zoom(self, widget, mode, adjustments=None):
		""" Sets the zoom level. """
		self.zoom_fit=False
		for widget in [self.tb_bestzoom, self.menu_bestzoom, self.tb_resetzoom, self.menu_resetzoom, self.tb_zoomin, self.menu_zoomin, self.tb_zoomout, self.menu_zoomout]:
			widget.set_sensitive(True)
		if self.canvas_geo.a%180==0:
			best=min(float(self.layout_width)/self.canvas_geo.w, float(self.layout_height)/self.canvas_geo.h)
		else:
			best=min(float(self.layout_width)/self.canvas_geo.h, float(self.layout_height)/self.canvas_geo.w)	
		if type(mode)==str:
			if mode=="+":
				self.canvas_geo.s*=1.25
				if self.canvas_geo.s*1.25>10:
					self.tb_zoomin.set_sensitive(False)
					self.menu_zoomin.set_sensitive(False)
			elif mode=="-":
				self.canvas_geo.s*=0.8
				if self.canvas_geo.s*0.8<0.1:
					self.tb_zoomout.set_sensitive(False)
					self.menu_zoomout.set_sensitive(False)
			if (self.canvas_geo.s/best>=0.90 and self.canvas_geo.s/best<=1.09) or mode=="f":
				self.zoom_fit=True
				self.tb_bestzoom.set_sensitive(False)
				self.menu_bestzoom.set_sensitive(False)
				self.canvas_geo.s=best
		elif type(mode)==int or type(mode)==float:
			self.canvas_geo.s=float(mode)
			if self.canvas_geo.s==1:
				self.tb_resetzoom.set_sensitive(False)
				self.menu_resetzoom.set_sensitive(False)
		if adjustments==None:
			if self.hadjustment.get_upper()-self.hadjustment.get_page_size()==0:
				xadj=0.5
			else:
				xadj=self.layout_scrollx/(self.hadjustment.get_upper()-self.hadjustment.get_page_size())
			if self.vadjustment.get_upper()-self.vadjustment.get_page_size()==0:
				yadj=0.5
			else:
				yadj=self.layout_scrolly/(self.vadjustment.get_upper()-self.vadjustment.get_page_size())
			adjustments=[xadj, yadj]
		self.canvas_geo.cw=int(round(self.canvas_geo.w*self.canvas_geo.s))
		self.canvas_geo.ch=int(round(self.canvas_geo.h*self.canvas_geo.s))
		if self.ssel!=None:
			self.ssel[0]=int(round(self.nsel[0]*self.canvas_geo.s))
			self.ssel[1]=int(round(self.nsel[1]*self.canvas_geo.s))
			self.ssel[2]=int(round(self.nsel[2]*self.canvas_geo.s))
			self.ssel[3]=int(round(self.nsel[3]*self.canvas_geo.s))
		if self.canvas_geo.a%180!=0:
			self.canvas_geo.cw, self.canvas_geo.ch = self.canvas_geo.ch, self.canvas_geo.cw
		self.canvas.set_size_request(self.canvas_geo.cw, self.canvas_geo.ch)
		self.__position_canvas(adjustments)
	
	def rotate(self, widget, angle):
		""" Rotates the canvas. """
		self.canvas_geo.a=(360+self.canvas_geo.a-angle)%360
		if self.ssel!=None:
			[w, h]=[self.canvas_geo.w, self.canvas_geo.h]
			if self.canvas_geo.a%180==0:
				w,h= h,w
			if angle==-90:
				self.ssel[0], self.ssel[1] = self.canvas_geo.ch-self.ssel[1]-self.ssel[3], self.ssel[0]
				self.nsel[0], self.nsel[1] = h-self.nsel[1]-self.nsel[3], self.nsel[0]
			elif angle==90:
				self.ssel[0], self.ssel[1] = self.ssel[1], self.canvas_geo.cw-self.ssel[0]-self.ssel[2]
				self.nsel[0], self.nsel[1] = self.nsel[1], w-self.nsel[0]-self.nsel[2]
			self.ssel[2],self.ssel[3]=self.ssel[3],self.ssel[2]
			self.nsel[2],self.nsel[3]=self.nsel[3],self.nsel[2]
		if self.zoom_fit:
			self.set_zoom(None,"f")
		else:
			self.canvas_geo.cw, self.canvas_geo.ch = self.canvas_geo.ch, self.canvas_geo.cw
			self.canvas.set_size_request(self.canvas_geo.cw, self.canvas_geo.ch)
			self.__position_canvas()
	
	def set_cursor(self, cursor):
		""" Sets the cursor over the drawingarea. """
		if cursor!=None:
			self.layout.window.set_cursor(cursor)
		elif self.image!=None:
			self.layout.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.CROSSHAIR))
		else:
			self.layout.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.LEFT_PTR))
	
	def __set_image_update_ui(self):
		""" Updates the UI after loading a new image (added through a gobject.idle_all
			to prevent a UI freeze since set_image is called in a different thread. """
		self.layout.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.CROSSHAIR))
		self.canvas.show()
		# Set spinbutton values
		self.spin_controls_brightness.handler_block_by_func(self.__brightness_contrast_changed)
		self.spin_controls_contrast.handler_block_by_func(self.__brightness_contrast_changed)
		self.spin_controls_brightness.set_value(0)
		self.spin_controls_contrast.set_value(0)
		self.spin_controls_brightness.handler_unblock_by_func(self.__brightness_contrast_changed)
		self.spin_controls_contrast.handler_unblock_by_func(self.__brightness_contrast_changed)
		self.spin_controls_resolution.handler_block_by_func(self.__resolution_changed)
		if self.document!=None:
			self.spin_controls_resolution.set_value(200)
			self.label_navigation_totpages.set_text('/ %d'%self.document.get_n_pages())
			self.spin_navigation_page.get_adjustment().set_upper(self.document.get_n_pages())
			if self.spin_navigation_page.get_value_as_int()!=1:
				self.spin_navigation_page.set_value(1)				# Callback renders the page
			else:
				self.__render_page(1)
			self.toolbar_navigation.show()
			self.label_controls_dpi.set_text("dpi")
		else:
			self.spin_controls_resolution.set_value(100)
			self.toolbar_navigation.hide()
			self.label_controls_dpi.set_text("%")
			self.set_zoom(None,"f")
		self.spin_controls_resolution.handler_unblock_by_func(self.__resolution_changed)
		self.ssel=self.nsel=None
		self.tb_doocr.set_label(_("Recognize all"))
		self.tb_doocr.set_tooltip_text(_("Recognize entire image"))
		return False
		
	def set_image(self, filename):
		""" Loads an image or a PDF document. """
		if os.path.splitext(filename)[1].lower()=='.pdf':
			try:
				self.document=poppler.document_new_from_file('file:%s'%urllib.pathname2url(filename), None)
			except:
				return False
		else:
			try:
				pixbuf=gtk.gdk.pixbuf_new_from_file(filename)
			except:
				return False
			self.document=None
			# Create surface
			self.image = cairo.ImageSurface(cairo.FORMAT_ARGB32, pixbuf.get_width(), pixbuf.get_height())
			context = cairo.Context(self.image)
			if pixbuf.get_has_alpha():										# Overlay transparent to white
				context.set_source_rgba(1.0, 1.0, 1.0, 1.0)
				context.paint()
			gdkcontext = gtk.gdk.CairoContext(context)
			gdkcontext.set_source_pixbuf(pixbuf, 0, 0)
			gdkcontext.paint()
			self.PILimage=Image.frombuffer("RGBA", (self.image.get_width(), self.image.get_height()), self.image.get_data(),"raw","RGBA", 0, 1).convert("RGB")
			self.origPILimage=self.PILimage.copy()
			# Update canvas state
			self.canvas_geo.w=self.image.get_width()
			self.canvas_geo.h=self.image.get_height()
			self.canvas_geo.a=0
		gobject.idle_add(self.__set_image_update_ui)
		return True
	
	def __render_page(self, pageno):
		""" Renders the specified page on the canvas. """
		scale=self.spin_controls_resolution.get_value()/72.0
		page=self.document.get_page(pageno-1)
		[w, h]=page.get_size()
		[w,h]=[int(math.floor(w*scale)), int(math.floor(h*scale))]
		# Create surface, render page
		self.image=cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
		context=cairo.Context(self.image)
		context.set_source_rgba(1.0, 1.0, 1.0, 1.0)
		context.paint()
		context.scale(scale, scale)
		page.render(context)
		self.PILimage=Image.frombuffer("RGBA", (self.image.get_width(), self.image.get_height()), self.image.get_data(),"raw","RGBA", 0, 1).convert("RGB")
		# Update canvas state
		if self.nsel:
			relscale=w/float(self.canvas_geo.w)
			self.nsel=[ int(round(item*relscale)) for item in self.nsel ]
		self.canvas_geo.w=w
		self.canvas_geo.h=h
		self.canvas_geo.a=0
		self.__set_brightness_contrast()
		self.set_zoom(None,"f")
		
	def clear_image(self):
		""" Clears the displayed image. """
		if self.image!=None:
			self.image=None
			self.canvas_geo=dotdict({'x':0,'y':0,'w':0,'h':0,'cw':0,'ch':0,'s':1.0,'a':0})
			self.layout.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.LEFT_PTR))
			self.ssel=self.nsel=None
			self.canvas.hide()
			self.__position_canvas()
	
	### SELECTION FUNCTIONS ###
	def __selection_start(self, widget, event):
		""" Called when the user initiates selection with the mouse. """
		if self.image!=None and self.wsel==None:	# Prevent double firing
			if self.ssel!=None:
				area=gtk.gdk.Rectangle(self.ssel[0]-1, self.ssel[1]-1, self.ssel[2]+2, self.ssel[3]+2)
				self.ssel=None
				self.__canvas_exposed(None, dotdict({'area': area}))
			self.wsel=[	int(round(min(max(0,event.x-self.canvas_geo.x),self.canvas_geo.cw))),
						int(round(min(max(0,event.y-self.canvas_geo.y),self.canvas_geo.ch))),
							0, 0]
			self.ssel=[self.wsel[0], self.wsel[1], self.wsel[2], self.wsel[3]]
			self.motionhandler=self.layout.connect("motion_notify_event",self.__select)
			self.releasehandler=self.layout.connect("button_release_event",self.__selection_end)
	
	def __scroll_by(self, dx, dy):
		""" Called when the user selects over the visible boundaries, initiating autoscrolling. """
		if not self.scroll_animation:
			return False
		oldW,oldH=self.wsel[2],self.wsel[3]
		self.wsel[2]+=dx
		self.wsel[3]+=dy
		self.ssel[0]=int(round(min(self.wsel[0], self.wsel[0]+self.wsel[2])))
		self.ssel[1]=int(round(min(self.wsel[1], self.wsel[1]+self.wsel[3])))
		self.ssel[2]=int(round(abs(self.wsel[2])))
		self.ssel[3]=int(round(abs(self.wsel[3])))
		# Damage area
		x=int(round(self.wsel[0]+min(oldW, self.wsel[2])))-2
		w=int(round(abs(oldW-self.wsel[2])))+4
		y=int(round(self.wsel[1]+min(oldH, self.wsel[3])))-2
		h=int(round(abs(oldH-self.wsel[3])))+4
		# Invalidate damaged areas (pure x, pure y, mixed)
		self.canvas.window.freeze_updates()
		self.canvas.window.invalidate_rect(gtk.gdk.Rectangle(self.ssel[0]-2, y, self.ssel[2]+4, h), False)
		self.canvas.window.invalidate_rect(gtk.gdk.Rectangle(x, self.ssel[1]-2, w, self.ssel[3]+4), False)
		self.canvas.window.invalidate_rect(gtk.gdk.Rectangle(x, y, w, h), False)
		self.hscrollbar.set_value(self.hadjustment.get_value()+dx)
		self.vscrollbar.set_value(self.vadjustment.get_value()+dy)
		self.canvas.window.thaw_updates()
		if self.scroll_animation:
			gobject.timeout_add(100, self.__scroll_by, dx, dy)
		return False
		
	def __select(self,widget,event):
		""" Called during mouse motion after a started selection. """
		xspeed=yspeed=0
		if event.x-self.layout_scrollx<=0 and self.layout_scrollx!=0:
			xspeed=event.x-self.layout_scrollx
		elif event.x-self.layout_scrollx>=self.layout_width and self.layout_scrollx!=self.hadjustment.get_upper()-self.hadjustment.get_page_size():
			xspeed=event.x-self.layout_scrollx-self.layout_width
		if event.y-self.layout_scrolly<=0 and self.layout_scrolly!=0:
			yspeed=event.y-self.layout_scrolly
		elif event.y-self.layout_scrolly>=self.layout_height and self.layout_scrolly!=self.vadjustment.get_upper()-self.vadjustment.get_page_size():
			yspeed=event.y-self.layout_scrolly-self.layout_height
		if xspeed==0 and yspeed==0:
			self.scroll_animation=False
		else:
			self.scroll_animation=True
			gobject.timeout_add(100,self.__scroll_by,xspeed/10.,yspeed/10.)
		oldW,oldH=self.wsel[2],self.wsel[3]
		self.wsel[2]=min(max(0,event.x-self.canvas_geo.x),self.canvas_geo.cw)-self.wsel[0]
		self.wsel[3]=min(max(0,event.y-self.canvas_geo.y),self.canvas_geo.ch)-self.wsel[1]
		self.ssel[0]=int(round(min(self.wsel[0], self.wsel[0]+self.wsel[2])))
		self.ssel[1]=int(round(min(self.wsel[1], self.wsel[1]+self.wsel[3])))
		self.ssel[2]=int(round(abs(self.wsel[2])))
		self.ssel[3]=int(round(abs(self.wsel[3])))
		# Damage area
		x=int(round(self.wsel[0]+min(oldW, self.wsel[2])))-1
		w=int(round(abs(oldW-self.wsel[2])))+2
		y=int(round(self.wsel[1]+min(oldH, self.wsel[3])))-1
		h=int(round(abs(oldH-self.wsel[3])))+2
		# Repaint damaged areas (pure x, pure y, mixed)
		self.canvas.window.freeze_updates()
		self.__canvas_exposed(None, dotdict({'area': gtk.gdk.Rectangle(x, self.ssel[1]-1, w, self.ssel[3]+2)}))
		self.__canvas_exposed(None, dotdict({'area': gtk.gdk.Rectangle(self.ssel[0]-1, y, self.ssel[2]+2, h)}))
		self.__canvas_exposed(None, dotdict({'area': gtk.gdk.Rectangle(x, y, w, h)}))
		self.canvas.window.thaw_updates()
	
	def __selection_end(self,widget,event):
		""" Called when the user releases the mouse button to stop selecting. """
		self.layout.disconnect(self.motionhandler)
		self.layout.disconnect(self.releasehandler)
		self.scroll_animation=False
		self.wsel=None
		if self.ssel[2]==0 or self.ssel[3]==0:
			self.tb_doocr.set_label(_("Recognize all"))
			self.tb_doocr.set_tooltip_text(_("Recognize entire image"))
			area=gtk.gdk.Rectangle(self.ssel[0]-1, self.ssel[1]-1, 2, 2)
			self.nsel=self.ssel=None
			self.__canvas_exposed(None, dotdict({'area': area}))
		else:
			self.tb_doocr.set_label(_("Recognize selection"))
			self.tb_doocr.set_tooltip_text(_("Recognize selected region"))
			self.nsel=[ int(round(value/self.canvas_geo.s)) for value in self.ssel]
	
	def __selection_draw(self, context, area):
		if self.ssel!=None:
			context.push_group()
			context.rectangle(self.ssel[0]+0.5, self.ssel[1]+0.5, self.ssel[2]-1, self.ssel[3]-1)
			context.set_source_rgba(self.selcolor[0],self.selcolor[1],self.selcolor[2], .25)
			context.fill()#_preserve()
			context.set_line_width(1.0)
			context.set_source_rgb(self.selcolor[0],self.selcolor[1],self.selcolor[2])
			context.move_to(self.ssel[0]+0.5, self.ssel[1]+0.5)
			context.rel_line_to(self.ssel[2]-1, 0.0)
			context.rel_line_to(0.0, self.ssel[3]-1)
			context.rel_line_to(-self.ssel[2]+1, 0.0)
			context.close_path()
			context.stroke()
			context.pop_group_to_source()
			sel=gtk.gdk.Rectangle(self.ssel[0]-1, self.ssel[1]-1, self.ssel[2]+2, self.ssel[3]+2)
			context.rectangle(area.intersect(sel))
			context.fill()
	
	def __selection_set_color(self,widget=None,event=None):
		""" Sets the selection color based on the GTK theme color scheme. """
		color=gtk.settings_get_default().get_properties("gtk-color-scheme")[0]
		i=color.find("selected_bg_color: ")
		if i==-1:
			self.selcolor=[0.0, 0.0, 1.0]
			return
		j=color[i:].find("\n")
		try:
			color=gtk.gdk.color_parse(color[i+19:i+j])
			self.selcolor=[color.red/65535.0,color.green/65535.0,color.blue/65535.0]
		except:
			self.selcolor=[0,0,1]
	
	def get_selection(self, page=None):
		""" Returns a PIL image object containing the selected part of the image of the specified page. If page is not specified, the current image is taken. """
		if page!=None: self.spin_navigation_page.set_value(page)
		if self.nsel==None:
			selection=Image.frombuffer("RGBA",(self.canvas_geo.w, self.canvas_geo.h), self.image.get_data(), "raw", "RGBA", 0, 1)
			if self.canvas_geo.a!=0: selection=selection.rotate(-self.canvas_geo.a)
		else:
			rw, rh=self.canvas_geo.w, self.canvas_geo.h
			if self.canvas_geo.a%180!=0: rw, rh=rh, rw
			surf=cairo.ImageSurface(cairo.FORMAT_ARGB32, self.nsel[2], self.nsel[3])
			context=cairo.Context(surf)
			context.translate(-self.nsel[0]+rw/2., -self.nsel[1]+rh/2.)
			context.rotate(self.canvas_geo.a/180.0*3.141592654)
			context.translate(-self.canvas_geo.w/2., -self.canvas_geo.h/2.)
			context.set_source_surface(self.image)
			context.paint()
			selection=Image.frombuffer("RGBA",(self.nsel[2], self.nsel[3]), surf.get_data(), "raw", "RGBA", 0, 1)
		return selection.convert("RGB")
	
	### Image Controls ###
	def __toggle_controls(self, widget):
		active=widget.get_active()
		if active:
			self.toolbar_controls.show()
		else:
			self.toolbar_controls.hide()
		self.menu_toggle_controls.handler_block_by_func(self.__toggle_controls)
		self.tb_toggle_controls.handler_block_by_func(self.__toggle_controls)
		self.menu_toggle_controls.set_active(active)
		self.tb_toggle_controls.set_active(active)
		self.menu_toggle_controls.handler_unblock_by_func(self.__toggle_controls)
		self.tb_toggle_controls.handler_unblock_by_func(self.__toggle_controls)
	
	def __brightness_contrast_changed(self, widget=None):
		""" Called when either the brightness or the contrast spin changes. """
		gobject.source_remove(self.controls_timer)
		self.controls_timer=gobject.timeout_add(100, self.__set_brightness_contrast)
		
	def __set_brightness_contrast(self):
		""" Changes the brightness and contrast of the image. """
		w, h=self.canvas_geo.w, self.canvas_geo.h
		im=ImageEnhance.Brightness(self.PILimage).enhance(1+self.spin_controls_brightness.get_value()/100)
		im=ImageEnhance.Contrast(im).enhance(1+self.spin_controls_contrast.get_value()/100)
		im=im.convert("RGBA")
		self.image=cairo.ImageSurface.create_for_data(array.array('B', im.tostring()), cairo.FORMAT_ARGB32, w, h, w*4)
		self.canvas.queue_draw()
		self.controls_timer=-1
		return False
	
	def __resolution_changed(self, widget):
		""" Called when the resolution spinbutton is changed. """
		gobject.source_remove(self.controls_timer)
		self.controls_timer=gobject.timeout_add(100, self.__change_resolution)
		
	def __change_resolution(self):
		""" Changes the resolution of the image. """
		if self.document!=None:
			self.__render_page(self.spin_navigation_page.get_value_as_int())
		else:
			scale=self.spin_controls_resolution.get_value()/100.
			newsize=[int(round(dim*scale)) for dim in self.origPILimage.size]
			self.PILimage=self.origPILimage.resize(newsize, Image.ANTIALIAS)
			if self.nsel:
				relscale=newsize[0]/float(self.canvas_geo.w)
				self.nsel=[ int(round(item*relscale)) for item in self.nsel ]
			self.canvas_geo.w, self.canvas_geo.h=newsize
			self.__set_brightness_contrast()
			self.set_zoom(None, "f")
		self.controls_timer=-1
		return False
	
	def __recognize(self, widget):
		if self.document!=None and self.document.get_n_pages()>1 and self.nsel==None:
			self.menu_doocr.popup(None, None, self.__doocr_menu_positioner, 1, gtk.get_current_event_time())
		else:
			self.main.recognize()
	
	def __doocr_menu_positioner(self, menu):
		x, y=self.main.window.window.get_origin()
		x+=self.tb_doocr.allocation.x
		y+=self.tb_doocr.allocation.y
		return (x, y, False)
	
	def __recognize_all(self, widget):
		self.main.recognize(range(1,self.document.get_n_pages()+1))
	
	def __recognize_pages(self, widget):
		self.entry_pages.set_text("")
		self.entry_pages.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse(self.main.get_config().get_color("base_color")))
		while True:
			response=self.dialog_pages.run()
			self.entry_pages.grab_focus()
			if response==gtk.RESPONSE_OK:
				pages=[]
				ranges=[ r.split('-') for r in re.sub(r'\s+','',self.entry_pages.get_text()).split(",")]
				try:
					for r in ranges:
						if len(r)==1:
							pages.append(int(r[0]))
						elif len(r)==2:
							pages+=range(int(r[0]),int(r[1])+1)
						else:
							raise
					pages=list(set(pages))
					pages.sort()
					N=self.document.get_n_pages()
					pages=filter(lambda x: x>0 and x<=N, pages)
					if pages: self.main.recognize(pages)	# If not empty
					break
				except:
					self.entry_pages.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('#FF8080'))
			else:
				break
		self.dialog_pages.hide()

