#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       notifier.py
#       
#       Copyright 2010-2011 Sandro Mani <manisandro@gmail.com>
#       
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Notifier classes:
#	Notifier: a bar appearing above the statusarea containing notifications
#	VersionNotifier: the class notifying about new versions, uses pynotify if available

import gtk
import os

havepynotify=False
try:
	import pynotify
	import webbrowser
	pynotify.init("gImageReader")
	havepynotify=True
except:
	pass

class Notifier:
	def __init__(self, builder):
		""" Initializes the notifier.
				builder (gtk.Builder): the gtkbuilder object from which gui objects are retreived.
		"""
		self.notifybox=builder.get_object("notifybox")
		self.notifylabel=builder.get_object("notifylabel")
		self.notifyhbox=builder.get_object("notifyhbox")
		self.closebtn=builder.get_object("btn_closenotify")
		self.notifybox.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("#FFE090"))
		builder.get_object("btn_closenotify").connect("clicked",self.hide)
		self.visible=False

	def hide(self, widget=None):
		""" Hides the notifier.
				widget=None (gtk.Widget): argument only used by signal callback
		"""
		if self.visible:
			self.notifybox.hide()
			self.visible=False
		
	def notify(self, title, text, actions=[]):
		""" Shows a notification.
				title (str): the title of the notification, shown bold
				text (str): the next of the notification 
				actions=[] (list): list of [name (str), callback (function), callback_args (list)]
				                   describing an action with specified name and callback. The signature
				                   of the callback must be function(widget, data, arg1, arg2, ...)
				                   where arg1, arg2, ... = callback_args
		"""
		for child in self.notifyhbox.get_children():
			if child!=self.closebtn:
				self.notifyhbox.remove(child)
		label=gtk.Label()
		label.set_markup('<b>%s:</b> %s' % (title, text))
		self.notifyhbox.pack_start(label, expand=False, fill=True, padding=2)
		label.show()
		for action in actions:
			ebox=gtk.EventBox()
			self.notifyhbox.pack_start(ebox, expand=False, fill=True, padding=4)
			ebox.show()
			label=gtk.Label()
			label.set_markup('<span color="blue">'+action[0]+'</span>')
			ebox.add(label)
			label.show()
			ebox.realize()
			ebox.connect("button_press_event", action[1], *action[2:])
			ebox.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND1))
			ebox.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("#FFE090"))
		self.notifybox.show()
		self.visible=True

if havepynotify:	#pynotify variant
	class VersionNotifier:
		def __init__(self, builder):
			""" Initializes the notifier.
					builder (gtk.Builder): the gtkbuilder object from which gui objects are retreived.
			"""
			self.notification=None
			self.curactions=[]
		
		def hide(self):
			""" Empty implementation to stay consistent with Notifier interface. """
			pass
		
		def notify(self, title, text, actions=[]):
			""" Shows a notification.
				title (str): the title of the notification, shown bold
				text (str): the next of the notification 
				actions=[] (list): list of [name (str), callback (function), callback_args (list)]
				                   describing an action with specified name and callback
			"""
			self.notification=pynotify.Notification(title, text, os.path.join(os.path.dirname(__file__),'gimagereader.png'))
			self.curactions=actions
			for action in actions:
				self.notification.add_action(str(actions.index(action)), action[0], self.__action_clicked)
			self.notification.show()
		
		def __action_clicked(self, notification, actionidx):
			i=int(actionidx)
			self.curactions[i][1](notification, None, *self.curactions[i][2:])

else:	#integrated variant
	VersionNotifier=Notifier


