#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       threadtools.py
#       
#       Copyright 2010-2011 Sandro Mani <manisandro@gmail.com>
#       
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.

import threading
import gtk
import gobject
import cairo
import math

# Prevent Popen from creating a terminal window on windoze
try:
	import win32process
	popencf=win32process.CREATE_NO_WINDOW
except:
	popencf=0

class SubprocessThread(threading.Thread):
	""" A class for handling threads. """
	def __init__(self, task, callback, *args):
		""" Initializes the new thread:
			 task: the function to call, accepting the (unpacked) arguments in args
			 callback: the function to call when the task completes, takes as parameter what was returned by task
			 args: variable number of input arguments to pass to the task function.
		"""	
		threading.Thread.__init__(self)
		self.setDaemon(True)
		self.task=task
		self.callback=callback
		self.aborted=False
		self.args=args
	
	def abort(self):
		""" Flags the thread to be terminated. """
		self.aborted=True
	
	def run(self):
		""" Runs the thread. Called by calling start() on the object. """
		if not self.args:
			result=self.task()
		else:
			result=self.task(*self.args)
		if self.aborted:
			return
		try:
			gobject.idle_add(self.callback, result)
		except:
			pass

class Spinner(gtk.DrawingArea):
	""" A specialized gtk.DrawingArea showing a spinner. """
	def __init__(self):
		""" Constructor. """
		gtk.DrawingArea.__init__(self)
		self.connect("expose-event", self.__redraw)
		self.connect("size-allocate", self.__resize)
		
		scheme=gtk.settings_get_default().get_properties("gtk-color-scheme")[0]
		if scheme=='':	# Fallback
			scheme='tooltip_fg_color: #000000000000\nselected_bg_color: #31316a6ac5c5\ntooltip_bg_color: #f5f5f5f5b5b5\nbase_color: #ffffffffffff\nfg_color: #000000000000\ntext_color: #000000000000\nselected_fg_color: #ffffffffffff\nbg_color: #efefececdcdc\n'
		scheme=dict(item.split(": ") for item in scheme.split("\n") if item!='')
		bg_color=gtk.gdk.color_parse(scheme['bg_color'])
		self.bg_color=(bg_color.red/65535.0, bg_color.green/65535.0, bg_color.blue/65535.0)

		self.ndots=10
		self.current=0
		self.offscreen=None
		self.offscreencontext=None
		self.show()
	
	def __del__(self):
		""" Destructor. """
		gobject.source_remove(self.timer)
	
	def __pulse(self):
		""" Updates the spinner by one "tick". """
		self.__redraw()
		self.current=(self.current+1)%self.ndots
		return True
	
	def __resize(self, widget, allocation):
		""" Called when the widget is resized. """
		self.offscreen=cairo.ImageSurface(cairo.FORMAT_ARGB32, allocation.width, allocation.height)
		self.offscreencontext=cairo.Context(self.offscreen)
		
	def __redraw(self, widget=None, event=None):
		""" Called when the widget is exposed or by self.__pulse. """
		if self.window==None or self.offscreen==None: return
		width=self.allocation.width
		height=self.allocation.height
		cx=width/2.0
		cy=height/2.0
		radius=min(width, height)*0.375
		dotradius=radius*0.266666667
		# Pain on offscreen buffer
		self.offscreencontext.set_source_rgb(*self.bg_color)
		self.offscreencontext.paint()
		step=2.0*math.pi/self.ndots
		for i in range(self.current, self.ndots+self.current):
			self.offscreencontext.set_source_rgba(0.6, 0.6, 0.6, (i-self.current)/float(self.ndots))
			self.offscreencontext.arc(cx+radius*math.cos(i*step), cy+radius*math.sin(i*step), dotradius, 0, 2*math.pi)
			self.offscreencontext.fill()
		# Transfer offscreen buffer to onscreen buffer
		context=self.window.cairo_create()
		context.set_source_surface(self.offscreen)
		context.paint()
	
	def stop(self):
		""" Stops the spinning animation. """
		gobject.source_remove(self.timer)
	
	def start(self):
		""" Starts the spinning animation. """
		self.timer=gobject.timeout_add(250, self.__pulse)
